unit PSListU;

{$IOCHECKS OFF}

interface

uses
  Libc,
  SysUtils, Types, Classes, Variants, QGraphics, QControls, QForms, QDialogs,
  QMenus, QTypes, QImgList, QActnList, QExtCtrls, QComCtrls, QStdCtrls;

type
  TfrmMain = class(TForm)
    imgsImages: TImageList;
    lstvProcessInfo: TListView;
    timRefresh: TTimer;
    alActions: TActionList;
    actKillProcess: TAction;
    actClose: TAction;
    actRefresh: TAction;
    actAllProcesses: TAction;
    actNonRootProcesses: TAction;
    actYourProcesses: TAction;
    actRunningProcesses: TAction;
    mnuMenuBar: TMainMenu;
    mniOptions: TMenuItem;
    mniOptionsKillProcess: TMenuItem;
    N1: TMenuItem;
    mniOptionsClose: TMenuItem;
    mniView: TMenuItem;
    mniViewRefresh: TMenuItem;
    N2: TMenuItem;
    mniAllProcesses: TMenuItem;
    mniYourProcesses: TMenuItem;
    mniNonRootProcesses: TMenuItem;
    mniRunningProcesses: TMenuItem;
    mniHelp: TMenuItem;
    mniHelpAbout: TMenuItem;
    mnuPopup: TPopupMenu;
    mniRefresh: TMenuItem;
    N3: TMenuItem;
    mniKillProcess: TMenuItem;
    actHelpAbout: TAction;
    StatusBar1: TStatusBar;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure actCloseExecute(Sender: TObject);
    procedure actRefreshExecute(Sender: TObject);
    procedure actViewProcessesExecute(Sender: TObject);
    procedure actKillProcessExecute(Sender: TObject);
    procedure actKillProcessUpdate(Sender: TObject);
    procedure actHelpAboutExecute(Sender: TObject);
  private
    //Some convenient user name strings
    CurrentUser, Root: String;
    //Directory of process being worked on (/proc/XXX)
    PIDDir: String;
    //PID of process being worked on
    PID: __pid_t;
    //Used to keep same item selected
    LastSelectedPID: String;
    //An easy to read version of /proc/XXX/status
    ProcessStatus: TStringList;
    procedure LoadTextFile(const Filename: String; List: TStrings);
    procedure GetProcessInfo;
    function GetProcessStatusLine(const Prefix: String): String;
    function GetProcessUser: String;
    function GetSymLinkPath(const SymLink: String): String;
    function GetProcessCmdLine: String;
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.xfm}

const
  colPID = 1;
{$if Declared(RTLVersion) and (RTLVersion < 14.20)}
  //Kylix 1 omission
  _POSIX_PATH_MAX = 1024;
{$ifend}


procedure TfrmMain.FormCreate(Sender: TObject);
begin
  //Default to safe file access mode
  FileMode := fmOpenRead or fmShareDenyNone;
  //Set up root & current user name strings
  Root := getpwuid(0)^.pw_name;
  CurrentUser := getpwuid(getuid)^.pw_name;
  //Set up string list for status files to be read into
  ProcessStatus := TStringList.Create;
  //Do first read of processes
  actAllProcesses.Checked := True;
  actAllProcesses.Execute;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  ProcessStatus.Free
end;

procedure TfrmMain.actCloseExecute(Sender: TObject);
begin
  Close
end;

//TStringList.LoadFromFile doesn't work on the
//kernel memory-mapped files as their size is reported as 0
procedure TfrmMain.LoadTextFile(const Filename: String; List: TStrings);
var
  F: TextFile;
  S: String;
  Stop: Boolean;
begin
  AssignFile(F, FileName);
  Reset(F);
  try
    List.Clear;
    repeat
      ReadLn(F, S);
      Stop := (IOResult <> 0) or (S = '');
      if not Stop then
        List.Add(S)
    until Stop
  finally
    CloseFile(F)
  end;
end;

//Locate a particular line from the status file, given a prefix string
function TfrmMain.GetProcessStatusLine(const Prefix: String): String;
var
  I: Integer;
  Line: String;
begin
  Result := '';
  for I := 0 to ProcessStatus.Count - 1 do
  begin
    Line := ProcessStatus[I];
    if Pos(Prefix, Line) > 0 then
    begin
      Result := Copy(Line, Succ(Length(Prefix)), Length(Line));
      Break;
    end
  end
end;

function TfrmMain.GetProcessUser: String;
var
  UID: string;
  TabPos: Integer;
begin
  UID := GetProcessStatusLine('Uid:'#9);
  TabPos := Pos(#9, UID);
  //Typically there are 3 UID values separated by tabs, but check, JIC
  if TabPos > 0 then
    Delete(UID, TabPos, Length(UID))
  else
    UID := Trim(UID);
  Result := getpwuid(StrToInt(UID))^.pw_name
end;

function TfrmMain.GetSymLinkPath(const SymLink: String): String;
var
  RealPathBuf: array[0.._POSIX_PATH_MAX] of Char;
begin
  realpath(PChar(SymLink), RealPathBuf);
  if SymLink <> String(RealPathBuf) then
    Result := RealPathBuf //if nothing doing, return N/A
  else
    Result := 'N/A'
end;

function TfrmMain.GetProcessCmdLine: String;
var
  F: file of Char;
  Ch: Char;
begin
  Result := '';
  //The file /proc/XXX/cmdline is a series of zero-terminated strings,
  //possibly followed by an extra zero terminator
  //Note that file size is reported as 0 for kernel memory-mapped files
  AssignFile(F, PIDDir + 'cmdline');
  Reset(F);
  try
    repeat
      Read(F, Ch);
      if Ch <> #0 then //read till a 0 char
        Result := Result + Ch
      else
      begin
        //We will find an extra #0 at EOF in some cases here
        //otherwise we'll trigger a read past EOF IOError
        Read(F, Ch);
        if (IOResult = 0) and (Ch <> #0) then
          Result := Result + #32 + Ch;
      end
    until (Ch = #0) or (IOResult <> 0)
  finally
    CloseFile(F)
  end
end;

procedure TfrmMain.GetProcessInfo;
var
  Item: TListItem;
  State, User: String;
begin
  LoadTextFile(PIDDir + 'status', ProcessStatus);
  //If user only wants running processes, leave for any other state
  State := GetProcessStatusLine('State:'#9);
  if actRunningProcesses.Checked and ((Length(State) = 0) or (State[1] <> 'R')) then
    Exit;
  //If user only wants non-root processes, leave if this is a root processes
  User := GetProcessUser;
  if actNonRootProcesses.Checked and (User = Root) then
    Exit;
  //If user only wants user processes, leave for other users' processes
  if actYourProcesses.Checked and (User <> CurrentUser) then
    Exit;
  //Add row to list view and fill in the columns.
  //Also try and re-select last selected process
  Item := lstvProcessInfo.Items.Add;
  Item.Caption := GetProcessStatusLine('Name:'#9);
  Item.SubItems.Add(GetSymLinkPath(PIDDir + 'exe'));
  Item.SubItems.Add(Format('%d (%:0x)', [PID]));
  if Item.SubItems[colPID] = LastSelectedPID then
    Item.Selected := True;
  Item.SubItems.Add(User);
  Item.SubItems.Add(State);
  Item.SubItems.Add(GetProcessStatusLine('VmSize:'#9));
  Item.SubItems.Add(GetProcessStatusLine('VmRSS:'#9));
  Item.SubItems.Add(GetSymLinkPath(PIDDir + 'cwd'));
  Item.SubItems.Add(GetProcessCmdLine);
end;

procedure TfrmMain.actKillProcessExecute(Sender: TObject);
begin
  kill(StrToInt(lstvProcessInfo.Selected.SubItems[colPID]), SIGKILL)
end;

procedure TfrmMain.actKillProcessUpdate(Sender: TObject);
begin
  (Sender as TAction).Enabled := Assigned(lstvProcessInfo.Selected)
end;

procedure TfrmMain.actRefreshExecute(Sender: TObject);
var
  SearchRec: TSearchRec;
begin
  lstvProcessInfo.Items.BeginUpdate;
  try
    if Assigned(lstvProcessInfo.Selected) then
      LastSelectedPID := lstvProcessInfo.Selected.SubItems[1]
    else
      LastSelectedPID := '';
    lstvProcessInfo.Items.Clear;
    if FindFirst('/proc/*', faAnyFile, SearchRec) = 0 then
      try
        repeat
          if SearchRec.Attr and faDirectory > 0 then
          begin
            PID := StrToIntDef(SearchRec.Name, 0);
            if PID <> 0 then
            begin
              PIDDir := Format('/proc/%s/', [SearchRec.Name]);
              GetProcessInfo;
            end
          end
        until FindNext(SearchRec) <> 0
      finally
        FindClose(SearchRec)
      end;
  finally
    lstvProcessInfo.Items.EndUpdate
  end
end;

procedure TfrmMain.actViewProcessesExecute(Sender: TObject);
begin
  actAllProcesses.Checked := False;
  actNonRootProcesses.Checked := False;
  actYourProcesses.Checked := False;
  actRunningProcesses.Checked := False;
  (Sender as TAction).Checked := True;
  actRefresh.Execute
end;

procedure TfrmMain.actHelpAboutExecute(Sender: TObject);
begin
  MessageDlg('Process Viewer'#10#10 +
    'Written in Borland Kylix'#174' Open Edition ' +
    'for Linux Format Magazine'#10#10 +
    'Copyright '#169' 2002 Brian Long (http://www.blong.com)',
    mtInformation, [mbOK], 0)
end;

end.
